#from response import HTTPClientResponse
from dez.buffer import Buffer
#more like a requester...


class HTTPClientWriter(object):
    def __init__(self, conn):
        self.conn = conn
        
    def dispatch(self, request, cb, args):
        request.headers['Host'] = self.conn.addr[0]
        self.conn.write(request.render(), self.__request_written_cb, [cb, args])
        
    def __request_written_cb(self, cb, args):
        return cb(*args)
        
#class HTTPClientRequest(object):
#    def __init__(self, writer):
#        self.writer = writer
        
class HTTPClientRequest(object):

    def __init__(self):
        self.protocol = "HTTP/1.1"
        self.method = "GET"
        self.path = "/"
        self.headers = {            
        }
        self.body = Buffer()
        
    def write(self, data):
        self.body += data
        self.headers['Content-Length'] = str(len(self.body))
        
    def render(self):
        output = "%s %s %s\r\n" % (self.method, self.path, self.protocol)
        output += "\r\n".join( [": ".join((key, val)) for (key, val) in self.headers.items() ])
        output += "\r\n\r\n"
        output += self.body.data
        return output
        
class HTTPClientRawRequest(object):

    def __init__(self):
        self.protocol = "HTTP/1.1"
        self.method = "GET"
        self.path = "/"
        
        pass




import rel, time
rel.override()
from dez.http.client import HTTPClient
from optparse import OptionParser

class LoadTester(object):
    def __init__(self, host, port, location, number, concurrency):
        self.url = "http://"+host+":"+port+location
        self.number = number
        self.concurrency = concurrency
        print "       server url:",self.url
        print "           number:",self.number
        print "      concurrency:",self.concurrency
        self.client = HTTPClient()
        self.client.client.start_connections(host, int(port), self.concurrency, self.connections_open, max_conn=concurrency)
        self.time_start = time.time()
        self.responses = 0

    def connections_open(self):
        self.time_request_start = self.time_request = time.time()
        print ""
        print "Running Load Test"
        print "   ", self.concurrency, 'connections opened in', int(1000*(self.time_request_start - self.time_start)), "ms"
        print "    -"
        for i in range(self.number):
            self.client.get_url(self.url, cb=self.response_cb, cbargs=[i+1])

    def response_cb(self, response, i):
        self.responses += 1
        if self.responses == self.number:
            now = time.time()
            print "   ", self.responses, 'responses:', int(1000*(now - self.time_request)), "ms"
            print "    -"
            print ""
            print "Requests Per Second"
            print "   ", self.number, 'requests handled in', int(1000*(now - self.time_request_start)), "ms"
            print "   ", int(self.number / (now - self.time_start)), "requests per second (with connection time)"
            print "   ", int(self.number / (now - self.time_request_start)), "request per second (without connection time)"
            return rel.abort()
        if not self.responses % 100:
            now = time.time()
            print "   ", self.responses, 'responses:', int(1000*(now - self.time_request))
            self.time_request = now

def parse_input():
    parser = OptionParser()
    parser.add_option("-d", "--domain", dest="domain", default="localhost", help="test server at DOMAIN")
    parser.add_option("-p", "--port", dest="port", default="80", help="test server at PORT")
    parser.add_option("-l", "--location", dest="location", default="/", help="path -> http:// DOMAIN : PORT LOCATION")
    parser.add_option("-n", "--number", dest="number", default="1000", help="total NUMBER of connections")
    parser.add_option("-c", "--concurrency", dest="concurrency", default="100", help="connection CONCURRENCY")
    return parser.parse_args()

def main():
    ops, args = parse_input()
    if not args:
        args = ["pyevent"]
    print ""
    print "Initializing Load Tester"
    print "   event listener:",rel.initialize(args)
    rel.signal(2, rel.abort)
    rel.timeout(30, rel.abort)
    l = LoadTester(ops.domain, ops.port, ops.location, int(ops.number), int(ops.concurrency))
    rel.dispatch()


from dez.network.server import SocketDaemon
from dez.http.server.router import Router
from dez.http.server.response import HTTPResponse
from request import HTTPRequest

class HTTPDaemon(object):
    def __init__(self, host, port, get_logger=None):
        self.__server = SocketDaemon(host, port, self.__new_connection)
        self.router = Router(self.__404, [])

    def __new_connection(self, conn):
        HTTPRequest(conn, self.router)

    def __404(self, request):
        r = HTTPResponse(request)
        r.status = "404 Not Found"
        r.write("The requested document %s was not found" % request.url)
        r.dispatch()

    def start(self):
        self.__server.start()

    def register_prefix(self, prefix, cb, args=[]):
        self.router.register_prefix(prefix, cb, args)



import socket
LQUEUE_SIZE = 5
BUFFER_SIZE = 4096

HTTP_DELIMITER = '\r\n'
HTTP_DELIM_LENGTH = len(HTTP_DELIMITER)
def server_socket(port):
    ''' Return a listening socket bound to the given interface and port. '''
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(('', port))
    sock.setblocking(0)
    sock.listen(LQUEUE_SIZE)
    return sock


def client_socket(addr, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setblocking(0)
    sock.connect_ex((addr, port))
    return sock    



from dez.http.errors import HTTPProtocolError

class HTTPRequest(object):
    def __init__(self, conn, router):
        self.__conn = conn
        self.__router = router
        self.headers = {}
        self.__pending = []
        self.__body_started = False
        self.__body_finished = False
        self.__head_finished = False
        self.__conn.set_rmode_delimiter("\r\n\r\n", self._process_head)

    def _process_head(self, data):
        print 'head',data
        self.__head_started = True
        self.action, headers = data.split("\r\n", 1)
        try:
            self.method, self.url, self.protocol = self.action.split(' ',2)
            url_scheme, version = self.protocol.split('/',1)
            self.version_major, self.version_minor = [int(half) for half in version.split('.',1)]
            self.url_scheme = url_scheme.lower()
        except ValueError:
            raise HTTPProtocolError, "Invalid HTTP status line"
        for header in headers.split("\r\n"):
            try:
                key, value = header.split(': ')
            except ValueError:
                raise HTTPProtocolError, "Invalid header:\r\n%s"%header
            self.headers[key.lower()] = value
        self.__content_length = int(self.headers.get('content-length','0'))
        dispatch_cb, args = self.__router(self.url)
        dispatch_cb(self, *args)
#        print 'finished head'
        self.__head_finished = True

    def _process_body(self, data):
        print 'body',data
        if not self.__head_finished or self.__body_finished:
            self.end()
            self.__pending.append(("body",data))
            return self.__next_action()
#        if not self.__head_finished:
#            raise HTTPProcessingError, "body cannot be processed before head"
#        if self.__body_finished:
#            raise HTTPProcessingError, "body already processed"
        self.__body_started = True
        self.__conn.set_rmode_size(self.__content_length, self.__body_cb, data)

    def _process_write(self, data):
        if self.__body_started and not self.__body_finished:
            self.__pending.append(("write",data))
            return self.__next_action()
        self.__body_finished = True
        self.__conn.write(*data)

    def __body_cb(self, data, cb, *args):
        cb(data, *args[0])
        self.__body_finished = True
        self.__next_action()

    def __write_cb(self, cb, args):
        if cb:
            cb(*args)
        self.__next_action()

    def __end_cb(self, cb):
        self.__init__(self.__conn, self.__router)

    def __close_cb(self, cb):
        self.__conn.close(cb)

    def __next_action(self):
        if self.__pending:
            action, data = self.__pending.pop(0)
            getattr(self, "_process_"+action)(data)

    def set_close_cb(self, cb, args):
        self.__conn.set_close_cb(cb, args)

    def read_body(self, cb=None, args=[]):
        self._process_body([cb, args])

    def read_body_stream(self, cb=None, args=[]):
        self.read_body(cb, args)

    def write(self, data, cb=None, args=[], eb=None, ebargs=[]):
        self._process_write([data, self.__write_cb, (cb, args), eb, ebargs])

    def end(self, cb=None):
        self.write("",self.__end_cb,[cb])

    def close(self, cb=None):
        self.write("",self.__close_cb,[cb])

"""
    def write(data, ...)
        if not self.read_body:
            return self.pending_action.append(('write', data, ...))
        self.conn.write(data, ..., ...)
    def read_body(self):
        if self.read_body or self.reading_body:
            raise DontReadTwice
        self.reading_body=True
        self.conn.set_mode_...(...)
    def body_complete(self):
        self.read_body = True
        ...
        for (action, ...) in self.pending_actions:
            if action == 'write':
                self._-write(...)
"""
